import os, sys
from pathlib import Path
sys.path.append(os.path.join(os.path.dirname(Path(__file__).parent),"src"))
from prometry import pdbloader as pl
from prometry import pdbgeometry as pg


import pandas as pd

DATADIR = "tests/data/"
ls_structures = ['6eex']

    
def test_one():
    ls_geos = ['N:CA+1','N:CA-1','N:CA+3']
    pobjs = []
    for pdb in ls_structures:            
        pla = pl.PdbLoader(pdb,DATADIR,cif=False,source="ebi")
        po = pla.load_pdb()
        pobjs.append(po)
    gm = pg.GeometryMaker(pobjs)
    df = gm.calculateGeometry(ls_geos)
    print(df)


if __name__ == "__main__":
    test_one()
    



from os.path import exists
import urllib.request

from Bio.PDB.MMCIFParser import MMCIFParser        
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.PDBParser import PDBParser
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from . import pdbobject as po

"""~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
class PdbLoader(object):
    def __init__(self, pdb_code, directory="", cif=False,source="ebi"):        
        self.pdb_code = pdb_code
        self.directory = directory        
        self.cif = cif
        self.pobj = po.PdbObject(pdb_code)                
        if source == "ebi":
            self.cif_filepath = f"{directory}{pdb_code}.cif"
            self.cif_url = f"https://www.ebi.ac.uk/pdbe/entry-files/download/{pdb_code}.cif"        
            self.pdb_filepath = f"{directory}{pdb_code}.pdb"
            self.pdb_url = f"https://www.ebi.ac.uk/pdbe/entry-files/download/pdb{pdb_code}.ent"
        elif source == "alphafold":
            #https://alphafold.ebi.ac.uk/files/AF-Q8CGX5-F1-model_v4.pdb
            #https://alphafold.ebi.ac.uk/files/AF-Q8CGX5-F1-model_v4.cif
            self.cif_filepath = f"{directory}{pdb_code}.cif"
            self.cif_url = f"https://alphafold.ebi.ac.uk/files/{pdb_code}.cif"
            self.pdb_filepath = f"{directory}{pdb_code}.pdb"
            self.pdb_url = f"https://alphafold.ebi.ac.uk/files/{pdb_code}.pdb"
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def exists_pdb(self):
        if exists(self._filepath):
            return True
        else:            
            return False
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def download_pdb(self,cif=False):
        if cif:
            if not exists(self.cif_filepath):
                try:            
                    urllib.request.urlretrieve(self.cif_url, self.cif_filepath)                                
                except:            
                    return False
        else:
            if not exists(self.pdb_filepath):
                try:            
                    urllib.request.urlretrieve(self.pdb_url, self.pdb_filepath)                                
                except:            
                    return False
        return True
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def load_pdb(self):
        loaded = False
        if self.cif:
            try:
                self.download_pdb(cif=True)
                structure = MMCIFParser().get_structure(self.pdb_code, self.cif_filepath)            
                loaded = True
            except Exception as e:                              
                print("Error loading cif file", str(e))
                loaded = False
        if not loaded:
            self.download_pdb(cif=False)            
            structure = PDBParser(PERMISSIVE=True).get_structure(self.pdb_code, self.pdb_filepath)                                        
        self.pobj.add_atoms(structure)                
        #print(structure.header)
        return self.pobj
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""



import math


def getMagnitude(x,y,z):
    return math.sqrt(x*x + y*y + z*z)

def getDistance(x1, y1, z1, x2, y2, z2):
    """
    param: x1,y1,z1= the coordinates of the first atom
    param: x2,y2,z2= the coordinates of the second atom
    returns: distance between atoms and details of the calculation
    """

    xd = float(x2) - float(x1)
    yd = float(y2) - float(y1)
    zd = float(z2) - float(z1)
    sumsqu = (xd * xd) + (yd * yd) + (zd * zd)
    mag = math.sqrt(sumsqu)
    #if mag > 10:
    #    print("mag",mag)
    return (mag)

def getAngle(x1, y1, z1, x2, y2, z2, x3, y3, z3):
    """
    param: x1,y1,z1= the coordinates of the first atom
    param: x2,y2,z2= the coordinates of the second atom
    param: x3,y3,z3= the coordinates of the third atom
    returns: angle between atoms and details of the calculation
    """

    xd = float(x2) - float(x1)
    yd = float(y2) - float(y1)
    zd = float(z2) - float(z1)
    xe = float(x2) - float(x3)
    ye = float(y2) - float(y3)
    ze = float(z2) - float(z3)
    dot = (xd * xe) + (yd * ye) + (zd * ze)
    magA = math.sqrt((xd * xd) + (yd * yd) + (zd * zd))
    magB = math.sqrt((xe * xe) + (ye * ye) + (ze * ze))
    cos_theta = dot / (magA * magB)
    theta = math.acos(cos_theta)
    theta_deg = (theta / 3.141592653589793238463) * 180
    return (round(theta_deg, 3))

def crossProduct(A, B):
    x = (A[1] * B[2]) - (A[2] * B[1])
    y = (A[2] * B[0]) - (A[0] * B[2])
    z = (A[0] * B[1]) - (A[1] * B[0])
    return (x, y, z)

def getDihedral(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4):
    """
    param: x1,y1,z1= the coordinates of the first atom
    param: x2,y2,z2= the coordinates of the second atom
    param: x3,y3,z3= the coordinates of the third atom
    param: x4,y4,z4= the coordinates of the fourth atom
    returns: angle between atoms and details of the calculation
    """

    xA = float(x2) - float(x1)
    yA = float(y2) - float(y1)
    zA = float(z2) - float(z1)

    xB = float(x2) - float(x3)
    yB = float(y2) - float(y3)
    zB = float(z2) - float(z3)

    xC = float(x4) - float(x3)
    yC = float(y4) - float(y3)
    zC = float(z4) - float(z3)

    crossAB = crossProduct((xA, yA, zA), (xB, yB, zB))
    crossBC = crossProduct((xB, yB, zB), (xC, yC, zC))
    dot = (crossAB[0] * crossBC[0]) + (crossAB[1] * crossBC[1]) + (crossAB[2] * crossBC[2])

    ABsq = (crossAB[0] ** 2) + (crossAB[1] ** 2) + (crossAB[2] ** 2)
    BCsq = (crossBC[0] ** 2) + (crossBC[1] ** 2) + (crossBC[2] ** 2)
    magAB = math.sqrt(ABsq)
    magBC = math.sqrt(BCsq)

    try:
        cos_theta = dot / (magAB * magBC)
        theta = math.acos(cos_theta)
        theta_deg = (theta / 3.141592653589793238463) * 180
        cross = crossProduct(crossAB, crossBC)
        dotB = (cross[0] * xB) + (cross[1] * yB) + (cross[2] * zB)
        if dotB > 0:
            theta_deg *= -1

        return (round(theta_deg, 3))
    except:
        return -180 #there is a 0 max for cos theta


import os, sys
from pathlib import Path
sys.path.append(os.path.join(os.path.dirname(Path(__file__).parent),"src"))
from prometry import pdbloader as pl
from prometry import pdbgeometry as pg


import pandas as pd

DATADIR = "tests/data/"
ls_structures = ['1ejg']

    
def test_aa():    
    ls_geos = ['N:CA[aa|GLY]']
    pobjs = []
    for pdb in ls_structures:            
        pla = pl.PdbLoader(pdb,DATADIR,cif=False,source="ebi")
        po = pla.load_pdb()
        pobjs.append(po)
    gm = pg.GeometryMaker(pobjs)
    df = gm.calculateGeometry(ls_geos)
    print(df.T)

def test_no():    
    ls_geos = ['N:{N,O}','N:{N,O@1}','N:{N,O&1}','N:{N,O&1}','N:(N,O)','N:(N,O@1)','N:(N,O&1)','N:(N,O&2)']
    #ls_geos = ['N:(N,O&1)']
    pobjs = []
    for pdb in ls_structures:            
        pla = pl.PdbLoader(pdb,DATADIR,cif=False,source="ebi")
        po = pla.load_pdb()
        pobjs.append(po)
    gm = pg.GeometryMaker(pobjs)
    df = gm.calculateGeometry(ls_geos)
    print(df.T)


if __name__ == "__main__":
    test_aa()
    test_no()
    


"""
RSA 4/2/23
https://pynative.com/make-python-class-json-serializable/#:~:text=Use%20toJSON()%20Method%20to%20make%20class%20JSON%20serializable&text=So%20we%20don't%20need,Python%20Object%20to%20JSON%20string.

"""

from leuci_xyz import vectorthree as v3
import pandas as pd

class PdbObject(object):
    def __init__(self, pdb_code):
        # PUBLIC INTERFACE        
        self.pdb_code = pdb_code
        self.resolution = -1
        self.exp_method = ""
        self.chains = {}
        self.exc_hetatm = False
              
    def __str__(self):
        return f"{self.pdb_code}\t{self.resolution}\t{self.exp_method}"
    
    def lines(self,atoms=False):
        lines = []
        for chain,resdic in self.chains.items():
            for no,res in resdic.items():                
                lines.append(chain + "\t" + str(res))
                for attype,atm in res.atoms.items():
                    if atoms:
                        lines.append(atm)
        return lines
                                
    def add_atoms(self,bio_struc):
        self.bio_struc = bio_struc
        self.resolution = self.bio_struc.header['resolution']        
        self.exp_method = self.bio_struc.header["structure_method"]                
        atomNo = 0
        ridx = 0
        last_bad = ""
        for model in self.bio_struc:
            for chain in model:
                self.chains[chain.id] = {} ####  a chain is a dictionary of residue number to GeoResidue ############################################################
                for residue in chain:
                    aa = residue.get_resname()
                    rid = residue.get_full_id()[3][1]
                    resd = PdbResidue(aa,rid,ridx)
                    chain = residue.get_full_id()[2]
                    hetatm = residue.get_full_id()[3][0]
                    aah = residue.get_full_id()
                    mutation = aah[3][2]                                                                                
                    #if rid == 59: #debug line
                    #    print(rid)
                    this_bad = chain + "_" + str(rid)
                    if this_bad != last_bad:                                    
                        if rid in self.chains[chain] and self.exc_hetatm:                          
                            # if it is already there something is wrong so lets remove it
                            del self.chains[chain][rid]
                        elif mutation[0] != " " and self.exc_hetatm:                          
                            # if there are mutation insertions they cause us geoemtry problems and we will remove them                        
                            last_bad = chain + "_" + str(rid)
                            if rid in self.chains[chain]: #we know this has alreadybeen checked, it is just a line of code to stop it adding
                                del self.chains[chain][rid]
                        else:                        
                            if str(hetatm[0:2]) == "H_" and self.exc_hetatm:
                                if False:
                                    print("HETATM",rid)
                            else:
                                # only proceed if we explicitly want hetatms                        
                                ridx = ridx + 1

                                for atom in residue:
                                    disordered = 'N'
                                    if atom.is_disordered():
                                        disordered = 'Y'
                                        if atom.disordered_has_id("A"):
                                            atom.disordered_select("A")
                                    if atom.get_occupancy() < 1:
                                        disordered = 'Y'
                                    atomNo += 1
                                    atom_name = atom.get_name()
                                    occupant = atom.get_full_id()[4][1]
                                    if occupant == ' ':
                                        occupant = 'A'
                                    x = atom.get_vector()[0]
                                    y = atom.get_vector()[1]
                                    z = atom.get_vector()[2]
                                    bfactor = atom.get_bfactor()
                                    occupancy = atom.get_occupancy()
                                    one_atom = PdbAtom(chain,resd,atom_name[0],atom_name,atomNo,disordered,occupancy,bfactor,x,y,z)
                                    resd.atoms[atom_name] = one_atom
                                self.chains[chain][rid] = resd
                                last_bad = ""

    def elementInList(self,element, atomlist):
        for atm in atomlist:
            if element in atm:
                return True, atm
        return False, None


    def dataFrame(self):        
        dicdfs = []        
        for chain,resdic in self.chains.items():
            for no,res in resdic.items():
                for attype,atm in res.atoms.items():     
                    dic={'pdbCode':self.pdb_code,'resolution':self.resolution,
                    'chain':atm.chain,'aa':res.amino_acid,'rid':res.rid,'ridx':res.ridx,                    
                    'atom':atm.atom_name, 'atomNo':atm.atom_no,'element':atm.atom_type,
                    'bfactor':atm.bfactor, 'occupancy':atm.occupancy,
                    'x':atm.x, 'y':atm.y, 'z':atm.z}
                    dicdfs.append(dic)
        return pd.DataFrame.from_dict(dicdfs)

    
###################################################################################################################
class PdbResidue:
    """Class for a single residue

    :data amino acid:
    :data atoms:
    """

    def __init__(self, amino_acid,rid,ridx):
        """Initialises a GeoPdb with a biopython structure

        :param biopython_structure: A list of structures that has been created from biopython
        """

        self.atoms = {}
        self.elements = {}
        self.amino_acid = amino_acid
        self.rid = rid
        self.ridx = ridx

    def __str__(self):
       delim = "\t"
       return f"{self.amino_acid}{delim}{self.rid}{delim}{self.ridx}"
    
    def infoResd(self):
       delim = "|"
       return f"{self.amino_acid}{delim}{self.rid}"

###################################################################################################################
class PdbAtom:
    """Class for a single atom

    :data atom_type:
    :data atom_name:
    :data x:
    :data y:
    :data z:
    :data disordered:
    :data occupancy:
    :data bfactor:
    """

    def __init__(self, chain,res,atom_type,atom_name,atom_no,disordered,occupancy,bfactor,x,y,z):
        """Initialises a pdb with a biopython structure

        :param biopython_structure: A list of structures that has been created from biopython
        """

        self.info = {}
        self.chain = chain
        self.res = res
        self.atom_type = atom_type
        self.atom_name = atom_name
        self.atom_no = atom_no
        self.disordered = disordered
        self.occupancy = occupancy
        self.bfactor = bfactor
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        delim = "\t"
        return f"{self.atom_type}{delim}{self.atom_name}{delim}{self.atom_no}{delim}{self.disordered}{delim}{self.occupancy}{delim}{self.bfactor}{delim}({self.x}{self.y}{self.z})"

    def infoAtom(self):
        delim = "|"
        return f"({self.chain}{delim}{self.res.infoResd()}{delim}{self.atom_name}{delim}{self.atom_no})"

    def matchesCriteria(self, criteria,dis=-1):
        if criteria == "":
            return True
        crits = criteria.split(",")
        for crit in crits:
            cri = crit.split("|")
            if cri[0].lower() == "aa":
                if self.res.amino_acid.upper() == cri[1].upper():
                    pass
                else:
                    return False
            if cri[0].lower() == "~aa":
                if self.res.amino_acid.upper() == cri[1].upper():
                    return False
                else:
                    pass
            elif cri[0].lower() == "dis" and dis != -1:
                if "><" in cri[1]:
                    crits = cri[1].split('>')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    if dis <= crit_dis1 and dis >= crit_dis2:
                        pass
                    else:
                        return False
                elif "<>" in cri[1]:
                    crits = cri[1].split('<')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    if dis >= crit_dis1 and dis <= crit_dis2:
                        pass
                    else:
                        return False
                elif "<" in cri[1]:
                    crit_dis = cri[1][1:]
                    if float(dis) <= float(crit_dis):
                        pass
                    else:
                        return False            
                elif ">" in cri[1]:
                    crit_dis = cri[1][1:]
                    if float(dis) >= float(crit_dis):
                        pass
                    else:
                        return False
            
            elif cri[0].lower() == "occ":
                if "=" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) == float(crit_occ):
                        pass
                    else:
                        return False            
                elif "<" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) <= float(crit_occ):
                        pass
                    else:
                        return False            
                elif ">" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) >= float(crit_occ):
                        pass
                    else:
                        return False
                
        return True
